import os
import logging
import datetime
import sys
from pathlib import Path

def get_logger(name: str = None):
    LOG_DIR = Path.home() / 'Logs' / 'welo365'
    LOG_DIR.mkdir(exist_ok=True)
    LOGGER_NAME = f"{name}.logger" if name else 'logger'
    logfile = os.path.join(LOG_DIR, f"{datetime.datetime.today().strftime('%Y%m%d')}.log")
    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(logging.DEBUG)
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    date_format = "%H:%M:%S"
    formatter = logging.Formatter(log_format, date_format)
    ch = logging.StreamHandler(sys.stderr)
    ch.setFormatter(formatter)
    ch.setLevel(logging.WARNING)
    logger.addHandler(ch)
    fh = logging.FileHandler(logfile)
    fh.setFormatter(formatter)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    return logger


import pandas as pd
from O365.drive import File
from typing import Union

from rypython.ry365 import WorkBook


def read_excel365(xl_file: Union[File, WorkBook], sheet_name: str = False, skip_rows: int = 0):
    wb = WorkBook(xl_file) if isinstance(xl_file, File) else xl_file
    if sheet_name is None:
        wss = wb.get_worksheets()
        dfs = {}
        for ws in wss:
            dfs[ws.name] = read_excel365(wb, sheet_name=ws.name, skip_rows=skip_rows)
        return dfs
    if sheet_name is False:
        wss = wb.get_worksheets()
        ws = wss[0]
    elif sheet_name:
        ws = wb.get_worksheet(sheet_name)
    _range = ws.get_used_range()
    cols, *values = _range.values[skip_rows:]
    df = pd.DataFrame(values, columns=cols)
    df.range = _range
    return df






from O365.sharepoint import Sharepoint as _Sharepoint
from O365.sharepoint import Site as _Site

from rypython.ry365.drive import Storage


class Site(_Site):
    def __init__(self, *, parent=None, con=None, **kwargs):
        super().__init__(parent=parent, con=con, **kwargs)
        self.site_storage = Storage(parent=self, main_resource=f"/sites/{self.object_id}")


class Sharepoint(_Sharepoint):
    site_constructor = Site

    def __init__(self, *, parent=None, con=None, **kwargs):
        super().__init__(parent=parent, con=con, **kwargs)



from __future__ import annotations

import re
import logging
import os
import requests
from requests.exceptions import HTTPError
# from rypython.ry365.logger import get_logger
from O365.excel import Range as _Range
from O365.excel import WorkBook as _WorkBook
from O365.excel import WorkSheet as _WorkSheet


logging.basicConfig(level=os.environ.get('LOGLEVEL', 'WARNING'))

class Range(_Range):
    pattern = r'^.*!(?P<left>[A-Z]+)(?P<top>[0-9]+)(:(?P<right>[A-Z]+)(?P<bottom>[0-9]+))?$'

    def __init__(self, parent=None, session=None, **kwargs):
        super().__init__(parent=parent, session=session, **kwargs)
        self.ws = parent
        self.matchgroup = re.search(self.pattern, self.address).groupdict()

    def batch_update(self, divs: int = 5):
        range_length = len(self.values)
        step = range_length // divs
        top = self.top
        left = self.left
        right = self.right
        batches = [(self.values[i:i+step], i+top) for i in range(0, len(self.values), step)]
        for batch, start_row in batches:
            update_address = f"{left}{start_row}:{right}{start_row + len(batch) - 1}"
            print(update_address)
            _range = self.ws.get_range(update_address)
            _range.update(values=batch, allow_batch=False)

    def update(self, values: list, divs: int = None, allow_batch: bool = True):
        self.values = values
        if allow_batch and (len(self.values) > 1000 or len(self.values[0]) > 1000):
            divs = divs or (len(self.values) // 1000) + 1
            return self.batch_update(divs=divs)
        try:
            super().update()
        except (requests.exceptions.HTTPError, HTTPError):
            logger.error(f"HTTP Error. Adjusting update shape.")
            max_len = max(len(row) for row in self.values)
            for row in self.values:
                for i in range(max_len - len(row)):
                    row.append('')
            self.batch_update()

    @property
    def left(self):
        return self.matchgroup.get('left')

    @property
    def right(self):
        return self.matchgroup.get('right')

    @property
    def top(self):
        return int(self.matchgroup.get('top'))

    @property
    def bottom(self):
        return int(self.matchgroup.get('bottom'))


class WorkSheet(_WorkSheet):
    range_constructor = Range

    def protect(self):
        payload = {
            'options': {
                'allowFormatCells': False,
                'allowFormatColumns': False,
                'allowFormatRows': False,
                'allowInsertColumns': False,
                'allowInsertRows': False,
                'allowInsertHyperlinks': False,
                'allowDeleteColumns': False,
                'allowDeleteRows': False,
                'allowSort': True,
                'allowAutoFilter': True,
                'allowPivotTables': True
            }
        }
        url = self.build_url('/protection/protect')
        return bool(self.session.post(url, json=payload))

    def unprotect(self):
        url = self.build_url('/protection/unprotect')
        return bool(self.session.post(url))

    def auto_fit_columns(self):
        _range = self.get_used_range()
        format = _range.get_format()
        format.auto_fit_columns()


class WorkBook(_WorkBook):
    worksheet_constructor = WorkSheet

    def __init__(self, file_item, *, use_session=False, persist=False):
        super().__init__(file_item, use_session=use_session, persist=persist)



from __future__ import annotations

import logging
import os
import re
from pathlib import Path
from urllib.parse import parse_qs, urlparse, unquote

import yaml
from O365 import Account, FileSystemTokenBackend
from O365.connection import MSGraphProtocol

from rypython.ry365.sharepoint import Sharepoint

logging.basicConfig(level=os.environ.get('LOGLEVEL', 'WARNING'))

RYPYTHON_CONFIG_PATH = os.environ.get(
    'RYPYTHON_CONFIG_PATH',
    Path.home() / '.config' / 'rypython'
)
if isinstance(RYPYTHON_CONFIG_PATH, str):
    RYPYTHON_CONFIG_PATH = Path(RYPYTHON_CONFIG_PATH)

if not RYPYTHON_CONFIG_PATH:
    raise MissingConfiguration('Missing local configuration path!')
RYPYTHON_CONFIG_FILE = RYPYTHON_CONFIG_PATH / 'config.yaml'
if not RYPYTHON_CONFIG_FILE:
    raise MissingConfiguration('Missing local configuration file!')

with open(RYPYTHON_CONFIG_FILE, 'r') as configfile:
    RYPYTHON_CONFIG = yaml.safe_load(configfile)

DOMAIN = RYPYTHON_CONFIG['sharepoint']['domain']
CLIENT_ID = RYPYTHON_CONFIG['sharepoint']['client_id']
CLIENT_SECRET = RYPYTHON_CONFIG['sharepoint']['client_secret']
CREDS = (CLIENT_ID, CLIENT_SECRET)
TOKEN_PATH = os.environ.get(
    'RYPYTHON_TOKEN_PATH',
    Path.home() / '.tokens'
)
if isinstance(TOKEN_PATH, str):
    TOKEN_PATH = Path(TOKEN_PATH)

'''
class TokenBackend(FileSystemTokenBackend):
    def save_token(self):
        """
        Saves the token dict in the specified file
        :return bool: Success / Failure
        """
        if self.token is None:
            raise ValueError('You have to set the "token" first.')

        try:
            if not self.token_path.parent.exists():
                self.token_path.parent.mkdir(parents=True)
        except Exception as e:
            logging.error('Token could not be saved: {}'.format(str(e)))
            return False

        with self.token_path.open('w') as token_file:
            # 'indent = True' will make the file human readable
            self.serializer.dump(self.token, token_file, indent=True)

        return True

'''


class O365Account(Account):
    def __init__(
            self,
            site: str = None,
            creds: tuple[str, str] = CREDS,
            scopes: list[str] = None,
            auth_flow_type: str = 'authorization'
    ):

        TOKEN = TOKEN_PATH / 'o365_token.txt'
        token_backend = None
        if TOKEN.exists():
            logging.debug(f"Using token file {TOKEN}")
            token_backend = FileSystemTokenBackend(token_path=TOKEN_PATH)
            token_backend.load_token()
            token_backend.get_token()
        scopes = scopes or ['offline_access', 'Sites.Manage.All', 'message_all']
        OPTIONS = {
            'token_backend': token_backend
        } if token_backend is not None else {
            'scopes': scopes,
            'auth_flow_type': auth_flow_type
        }
        super().__init__(creds, **OPTIONS)
        if not self.is_authenticated:
            self.authenticate()
        self.drives = self.storage().get_drives()
        self.site = self.get_site(site) if site else None
        self.drive = self.site.get_default_document_library() if self.site else self.storage().get_default_drive()
        self.root_folder = self.drive.get_root_folder()

    def sharepoint(self, *, resource=''):
        if not isinstance(self.protocol, MSGraphProtocol):
            raise RuntimeError(
                'Sharepoint api only works on Microsoft Graph API'
            )
        return Sharepoint(parent=self, main_resource=resource)

    def get_site(self, site: str):
        return self.sharepoint().get_site(DOMAIN, f"/sites/{site}")

    def search(self, query: str):
        u = urlparse(query)
        site = None
        drive = None
        file_name = None
        if (site_query := re.search(r'.*/sites/(?P<site>[\w_\-\d]+)/', u.path)):
            if (site := site_query.group('site')):
                site = self.get_site(site)
                drive = site.get_default_document_library()
        if (q := parse_qs(u.query)):
            file_name = q.get('file', [''])[0]
        drive = drive or self.drive
        query = file_name or query
        results = list(drive.search(query))
        return results[0] if results else results

    def get_folder(self, *subfolders: str, site: str = None):
        if len(subfolders) == 0:
            return self.drive

        site = self.get_site(site) if site else self.site
        drive = site.get_default_document_library() if site else self.drive

        items = drive.get_items()
        for subfolder in subfolders:
            try:
                subfolder_drive = list(filter(lambda x: subfolder in x.name, items))[0]
                items = subfolder_drive.get_items()
            except:
                raise ('Path {} not exist.'.format('/'.join(subfolders)))
        return subfolder_drive

    @staticmethod
    def get_item_by_url(item_path: str, library_map: dict = None):
        pattern = r'^.*\/sites\/(?P<site>[0-9.\-A-Za-z]+)\/(?:[0-9.\-A-Za-z\s]+)\/(?P<folder_path>.*)$'
        site, lib_name, *item_path = unquote(item_path).split('/')[4:]
        item_path = '/'.join(item_path)
        account = O365Account(site=site)
        doc_library = None
        if lib_name == 'Shared Documents':
            doc_library = account.site.get_default_document_library()
        if lib_name != 'Shared Documents':
            if library_map:
                lib_name = library_map.get(lib_name, lib_name)
            doc_library = account.get_document_library_by_name(lib_name)
        if not doc_library:
            raise ValueError(f"Could not find document library named {lib_name} on site {site}!")
        return doc_library.get_item_by_path(item_path)

    def get_folder_by_path(self, folder_path: str):
        pattern = r'^.*\/sites\/(?P<site>[0-9.\-A-Za-z]+)\/(?:[0-9.\-A-Za-z\s]+)\/(?P<folder_path>.*)$'
        folder_path = unquote(folder_path)
        matches = re.search(pattern, folder_path)
        if not matches:
            raise ValueError(f"Folder not found. Check that your path.")
        matches = matches.groupdict()
        site = matches.get('site')
        folder_path = matches.get('folder_path')
        if not site or not folder_path:
            raise ValueError(f"Folder not found. Check that your path.")
        return self.get_folder(
            *folder_path.split('/'),
            site=site
        )

    def get_folder_by_url(self, url: str, library_map: dict = None):
        site, lib_name, *folder_path = unquote(url).split('/')[4:]
        folder_path = '/'.join(folder_path)
        if site != self.site.name:
            return O365Account(site=site).get_folder_by_url(
                url, library_map=library_map
            )
        if lib_name != 'Shared Documents' and library_map:
            lib_name = library_map.get(lib_name)
        doc_library = self.get_document_library_by_name(
            lib_name
        ) if lib_name != 'Shared Documents' else self.site.get_default_document_library()
        return doc_library.get_item_by_path(folder_path)

    def get_document_library_by_name(self, document_library_name: str, site: str = None):
        site = self.get_site(site) if site else self.site
        for drive in site.list_document_libraries():
            if drive.name.lower() == document_library_name.lower():
                return drive



import re

def stringify(q: str):
    return q if isinstance(q, str) else str(q)

def capture_all(pattern: str, query: str, allow_empty: bool = True):
    query = stringify(query)
    r = re.compile(pattern)
    s = [
        m.groupdict()
        for m in r.finditer(query)
    ]
    d = {}
    for match in s:
        for key, value in match.items():
            if not value and not allow_empty:
                continue
            d.setdefault(key, []).append(value)
    return d

def is_format(pattern: str, query: str):
    query = stringify(query)
    r = re.compile(pattern)
    return bool(re.match(r, query))


from datetime import date

class Today(object):
    def __init__(self):
        self.today = date.today()

    def __call__(self, strftime: str):
        return self.today.strftime(strftime)


today = Today()




import logging
import requests
import os

from O365.drive import Drive as _Drive
from O365.drive import Folder as _Folder
from O365.drive import Storage as _Storage
from O365.drive import Image, Photo, File


logging.basicConfig(level=os.environ.get('LOGLEVEL', 'WARNING'))


class Folder(_Folder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_item(self, item_name: str):
        for item in self.get_items():
            if item_name.lower() in item.name.lower():
                return item

    @staticmethod
    def _classifier(item):
        if 'folder' in item:
            return Folder
        elif 'image' in item:
            return Image
        elif 'photo' in item:
            return Photo
        else:
            return File


class Drive(_Drive):
    def __init__(self, *, parent=None, con=None, **kwargs):
        super().__init__(parent=parent, con=con, **kwargs)

    @staticmethod
    def _classifier(item):
        if 'folder' in item:
            return Folder
        elif 'image' in item:
            return Image
        elif 'photo' in item:
            return Photo
        else:
            return File

    def get_item_by_path(self, *parts: str):
        item_path = f"/{'/'.join(parts)}"
        if self.object_id:
            url = self.build_url(
                self._endpoints.get('get_item_by_path').format(id=self.object_id,
                                                               item_path=item_path))
        else:
            url = self.build_url(
                self._endpoints.get('get_item_by_path_default').format(item_path=item_path))
        try:
            response = self.con.get(url)
            if not response:
                return None
            data = response.json()
            return self._classifier(data)(
                parent=self,
                **{self._cloud_data_key: data}
            )
        except requests.exceptions.HTTPError:
            return

class Storage(_Storage):
    drive_constructor = Drive

    def __init__(self, *, parent=None, con=None, **kwargs):
        super().__init__(parent=parent, con=con, **kwargs)

    def get_default_drive(self, request_drive=False):
        if request_drive is False:
            return Drive(con=self.con, protocol=self.protocol,
                         main_resource=self.main_resource, name='Default Drive')
        super().get_default_drive(request_drive=request_drive)



from pathlib import Path
from typing import Tuple, Callable
from xlsxwriter.format import Format

import pandas as pd
import xlsxwriter


class RexcelWorkbook:
    COLS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

    def __init__(
            self,
            output_file: Path
    ) -> None:
        self.worksheets = []
        self.output_file = output_file

    def __enter__(self):
        self.workbook = xlsxwriter.Workbook(self.output_file)
        return self

    def __exit__(self, type, value, traceback):
        self.workbook.close()

    def add_format(self, config: dict):
        return self.workbook.add_format(config)

    def add_worksheet_by_dataframe(
            self,
            df: pd.DataFrame,
            worksheet_name: str,
            column_widths: list,
            include_index: bool = False,
            format_rows: Tuple[Callable, Format] = None,
            formula_columns: dict = None,
            conditional_formatting: dict = None
    ):
        df = df.where(pd.notnull(df), '')
        format_test, row_format = format_rows if format_rows else (None, None)
        if include_index:
            df = df.reset_index()
        wks = self.workbook.add_worksheet(name=worksheet_name)
        bold = self.workbook.add_format({'bold': 1})
        if column_widths:
            for first, last, width in column_widths:
                wks.set_column(first, last, width)
        columns = df.columns.tolist()
        if formula_columns is not None:
            for formula_column in formula_columns:
                columns.append(formula_column)
        for j, header in enumerate(columns):
            if not header or header == 'index':
                continue
            wks.write(
                f"{self.COLS[j]}1",
                header,
                bold
            )
        rows = df.values.tolist()
        row_number = 1
        col_number = 0
        write_funcs = {
            str: wks.write_string,
            bool: wks.write_boolean,
            int: wks.write_number
        }
        for row in rows:
            offset = 0
            for j, cell in enumerate(row):
                if type(cell) not in write_funcs:
                    cell = str(cell)
                cell_info = [
                    row_number,
                    col_number + offset,
                    cell
                ]
                if format_rows is not None and format_test(row_number):
                    cell_info.append(row_format)
                write_func = write_funcs.get(type(cell))
                write_func(
                    *cell_info
                )
                offset += 1
            if formula_columns is not None:
                for row_test, formula_format in formula_columns.values():
                    if row_test(row):
                        cell_info = [
                            row_number,
                            col_number + offset,
                            formula_format(row, row_number)
                        ]
                        wks.write_formula(*cell_info)
                    offset += 1
            row_number += 1
        if conditional_formatting:
            for format_range, config in conditional_formatting.items():
                if '{end}' in format_range:
                    format_range = format_range.format(end=len(rows) + 1)
                wks.conditional_format(
                    format_range,
                    config
                )
        self.worksheets.append(wks)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rypython',
 'rypython.randas',
 'rypython.rex',
 'rypython.rexcel',
 'rypython.ry365',
 'rypython.rytime']

package_data = \
{'': ['*']}

install_requires = \
['O365>=2.0.14,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'XlsxWriter>=3.0.2,<4.0.0',
 'pandas>=1.2.1,<2.0.0']

setup_kwargs = {
    'name': 'rypython',
    'version': '1.3.10',
    'description': 'Miscellaneous python tools',
    'long_description': None,
    'author': "Ryan O'Rourke",
    'author_email': 'ryan@rypy.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
